% GPE Tuning example
% this example uses synthetic data

clear all;

% standard time set up
M=512;
Dt=1.0;
tmax=Dt*(M-1); 
t=Dt*[0:M-1]'; 

% standard frequency set up
fmax=1/(2.0*Dt);
df=fmax/(M/2);
f=df*[0:M/2,-M/2+1:-1]'; 
Nf=M/2+1;
dw=2*pi*df;
w=dw*[0:M/2,-M/2+1:-1]';
Nw=Nf;

% PART 1: create true covariance function
n=12;
w0 = n*2*pi/tmax;
tmid=tmax/2;
s = 20/tmax;
gamma = 1.0;
gamma2 = gamma^2;
c = gamma2*cos(w0*(t-tmid)).*exp(-s*abs(t-tmid));

% amplitude spectral density of true covariance
ct = fft(c);
asd = abs(ct);

% PART 2: create synthetic data with this covariance

% realize a time series d with this covariance 
d = random('normal',0,1,M,1); % random numbers
d = d - mean(d); % remove mean
dt = fft(d);     % Fourier transform of d
dasd = abs(dt);  % a.s.d. of d
dt = asd.*dt./dasd; % adjust so has a.s.d. of c
dtrue = real(ifft(dt)); % inverse Fourier transform
vard = var(dtrue);      % variance of dtrue
dtrue = gamma*dtrue/sqrt(vard); % normalize d to variance gamma^2
dasd = abs(fft(dtrue)); % corrected a.s.d of dtrue

% prior model parameters
mpri = zeros(M,1);

% time series figure
figure(1);
clf;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, M, -5, 5] );
xlabel('time t (s)');
ylabel('d(t) and  m(t)');
hold on;
plot(t,dtrue,'k-','LineWidth',4);

% PART 3: Create observed data by sub-sampling true
% data and adding random noise

% noise level was 0.5
sigma = 0.1;
sigma2 = sigma^2;
sigmar = 1/sigma;

N = 150; % fewer observed data than true data
j = unidrnd( M, N, 1 ); % randomly select times
j = sort(j,'ascend'); % sort for plotting purposes
tobs = t(j); % observed times
% observed data is true data + uncorrelated random noise
dobs = dtrue(j) + random('Normal', 0, sigma, N, 1 );

% selects data d from model paramerts m
G = zeros(N,M);
for i=[1:N]
    G(i,j(i)) = 1;
end

% plot observed data
figure(1);
plot(tobs,dobs,'ko','LineWidth',2);

% assign w0 and s used in GPR
w0est = 4.0*w0; % w0est = 1.0*w0 is the true value
sest = 1.0*s;

% Build covariance C(control,control)
C1 = zeros(N,N);
for i=[1:N]
for j=[1:N]
    C1(i,j) = gamma2*cos(w0est*(tobs(i)-tobs(j))) .* exp( -sest*abs(tobs(i)-tobs(j)) );
end
end

% Build covariance C(target,control)
C2 = zeros(M,N);
for i=[1:M]
for j=[1:N]
    C2(i,j) = gamma2*cos(w0est*(t(i)-tobs(j))) .* exp( -sest*abs(t(i)-tobs(j)) );
end
end

size(N)

% solve GPE problem
MM = (C1 + sigma2*eye(N,N) );
mest = C2 * (MM\dobs);
plot(t,mest,'r-','LineWidth',3);


% Newton's Method adjustment of p

xd = tobs;
d = dobs;
xm = t;
rr = s;
ptrue=w0;
pg = w0est;
Dp = 0.1;
NITER = 30;
pg = pg-Dp;  % so first iteration starts pg
alpha = 1.0;
nu = 1.0;
for iter = [1:NITER]
    pg = pg+nu*alpha*Dp; % increment
    
    A = zeros(N,N);
    Cdd = zeros(N,N);
    dAdp = zeros(N,N);
    dCdddp = zeros(N,N);
    pg2 = pg^2;
    for i=[1:N]
    for j=[1:N]
        Cdd(i,j) = gamma2 * cos( pg*(xd(i)-xd(j)) ) * exp( -rr*abs(xd(i)-xd(j)) );
        A(i,j) = Cdd(i,j) + sigma2*(i==j);
        dCdddp(i,j) = -gamma2 * (xd(i)-xd(j)) * sin(pg*(xd(i)-xd(j))) * exp(-rr*abs(xd(i)-xd(j)));
        dAdp(i,j) = dCdddp(i,j);
    end
    end

    Cmd = zeros(M,N);
    dCmddp = zeros(M,N);
    for i=[1:M]
    for j=[1:N]
        Cmd(i,j) = gamma2 * cos( pg*(xm(i)-xd(j)) ) * exp( -rr*abs(xm(i)-xd(j)) );
        dCmddp(i,j) = -gamma2 * (xm(i)-xd(j)) * sin(pg*(xm(i)-xd(j))) * exp(-rr*abs(xm(i)-xd(j)));
    end
    end
    
    Cmm = zeros(M,M);
    dCmmdp = zeros(M,M);
    for i=[1:M]
    for j=[1:M]
        Cmm(i,j) = gamma2 * cos( pg*(xm(i)-xm(j)) ) * exp( -rr*abs(xm(i)-xm(j)) );
        dCmmdp(i,j) = -gamma2 * (xm(i)-xm(j)) * sin(pg*(xm(i)-xm(j))) * exp(-rr*abs(xm(i)-xm(j)));
    end
    end
    CmmI = inv(Cmm);
    CmmIsr = sqrtm(CmmI);
    dCmmIdp = - CmmI * dCmmdp * CmmI;
    dCmmIsrdp = sylvester( CmmIsr, CmmIsr, dCmmIdp );
    R = chol(Cmm);
    logdetCmm = sum(log(diag(R)));
    dlogdetCmmdp = trace( CmmI*dCmmdp );

    % solution
    u = A\d;
    mg = Cmd*u;
    dg = G*mg;

    % derivative of predicted data
    v = A\(dAdp*u);
    dmestdp = dCmddp*u - Cmd*v;
    ddpredp = G*dCmddp*u - G*Cmd*v;
    detildedp = -sigmar*ddpredp;
    e = d-dg;
    etilde = sigmar*e;
    
    % derivative of total data error
    E = etilde'*etilde;
    dEdp = 2*etilde'*detildedp;
    
    % total data error and its derivative
    el=(mpri-mg);
    eltilde = CmmIsr*el;
    deltildedq = dCmmIsrdp*el - CmmIsr*dmestdp;
    
    % total prior error and its derivative
    L = eltilde'*eltilde;
    dLdp = 2*eltilde'*deltildedq;
    
    PSI = logdetCmm + E + L;
    dPSIdp = dlogdetCmmdp + dEdp + dLdp;
    nu = -sign(dPSIdp);  % downhill vector
   
    % gradient method method
    if( iter==1 )  % on the first iteration do nothing but save
        pglast = pg;
        PSIlast = PSI;
        fprintf('1: pg %e nu %e alpha %e PSI %e PSIlast %e\n', pg, nu, alpha, PSI, PSIlast);
    elseif (PSI < PSIlast )
        pglast = pg;
        PSIlast = PSI;
        fprintf('2: pg %e nu %e alpha %e PSI %e PSilast %e\n', pg, nu, alpha, PSI, PSIlast);
    else % psi increased, so decrease alpha
        pg = pglast;
        PSI = PSIlast;
        alpha = alpha/2;
        fprintf('2: pg %e nu %e alpha %e PSI %e PSIlast %e\n', pg, nu, alpha, PSI, PSIlast);
    end
        
end
pest = pg;

figure(1);
plot( xm, mg, 'g-', 'LineWidth', 2);

fprintf('ptrue %e pest %e relating error %e\n', ptrue, pest, (ptrue-pest)/ptrue );

