% edama_09_??, Gaussian Process Regression example
% this example uses synthetic data
% and a gamma2*cos(w0T)*exp(-sT) covariance function,
% with T = |t(i)-t(j)|
clear all;

% tunable parameters
MPLOT = 5;    % scale of model parameter plot
CPLOT = 1;    % scale of covariance plot
gamma = 1;    % sqrt(variance) of covariance
sigmad = 0.2; % noise level of observed data
wbias=1.0;    % compute GPR with wrong w0
sbias=1.0;    % compute GPR with wrong s
PLOTCONF=1;   % plot confidence intervals on 1
USEFORRLOOP=0; % for loops on 1, outer products on 0

% axes definitons

% model parameter axis, tm, with sampling Dt
M=256;  % must be even, best if a power of 2
Dtm=1.0;
tmax=Dtm*(M-1); 
tm=Dtm*[0:M-1]'; 

% finely sampled model parameter vector, with sampling
% Dt/K, where K is a smallish power of 2, is used in
% the initialization process, as described below
K=8; % must be even, best if a power of 2
Mf=M*K;  
Dtf=Dtm/K;
tf=Dtf*[0:Mf-1]';

% Part 1: realization of a finely sampled time
% series, mf, with the desired covariance

% PART 1: create true "fine" covariance function

n=6;  % number of oscillations on (0,tmax) interal
w0 = n*2*pi/tmax;  % angular frequency of oscilation
tmid=tmax/2.0;
s = 20.0/tmax;  % decay rate
gamma2 = gamma^2;
cf = gamma2*cos(w0*(tf-tmid)).*exp(-s*abs(tf-tmid));

% plot this covariance
figure(1);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ -tmid, tmid, -CPLOT, CPLOT] );
xlabel('time t (s)');
ylabel('c(t)');
plot(tf-tmid,cf,'k-','LineWidth',3);

% PART 2: create "fine" time series with thus
% covariance using Fourier techniques

% amplitude spectral density of true covariance
cftilde = fft(cf);
cfasd = abs(cftilde);

% To create a time series with a specified covariannce,
% start with a times series of uniformly-distrbuted
% random noise, which has a uniform a.s.d., Fourier
% transorm it, multiply teh transform by the a.s.d
% of the specified covariance, trasnform back, and adjust
% the amplitude to the desired variance.
mf = random('Normal',0.0,1.0,Mf,1); % random numbers
mf = mf - mean(mf); % remove mean
mftilde = fft(mf);      % Fourier transform of mf
mfasd = abs(mftilde);   % a.s.d. of mf
mftilde = (cfasd./mfasd).*mftilde;  % adjust so has a.s.d. of c
mf = real(ifft(mftilde)); % inverse Fourier transform
varmf = var(mf);      % variance of mf
mf = (gamma/sqrt(varmf))*mf; % normalize mf to variance gamma^2

% PART 3: Regularly sub-sample mf to get mtrue, irregularly
% sub-sample mf to get true data dtrue, and add uncorrelated
% noise to dtrue to get observed data dobs

mtrue = mf(1:K:Mf);

N = 50; % fewer observed data than true data
% create list, j, of N unique, randomly chosen indices
% that are not in the sequence 1, K+1, 2K+1, ...
j=zeros(N,1);
Nj=0;
while (Nj<N)
    i=unidrnd(Mf);
    if( (mod(i-1,K)~=0) && (~ismember(i,j)) )
        Nj=Nj+1;
        j(Nj)=i;
    end
end
j = sort(j,'ascend'); % sort for plotting purposes
td = tf(j); % times of observations
dtrue = mf(j);  % irregularly sub-sample mf to get true data
% observed data is true data + uncorrelated random noise
sigmad2 = sigmad^2;
dobs = dtrue + random('Normal', 0, sigmad, N, 1 );

% the data kernel, G, for the equation d=Gm
G = [ zeros(N,M), zeros(N,N) ];

% time series figure
figure(2);
clf;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, tmax, -MPLOT, MPLOT] );
xlabel('time t (s)');
ylabel('d(t) and  m(t)');
hold on;
plot(tm,mtrue,'k-','LineWidth',3);
plot(td,dtrue,'ko','LineWidth',2);

% Since the true (w0, s) are in general unknown
% this code allows them to be assigned differently
% from the true values
w0est = wbias*w0;  % w0est = 1.0*w0 is the true value
sest = sbias*s;

if( USEFORRLOOP )
    % build covariances via for loop
    Ccc = zeros(N,N); % covariance matrix, C(control,control)
    for i=[1:N]
    for j=[1:N]
        T = abs(td(i)-td(j));
        Ccc(i,j) = gamma2*cos(w0est*T) * exp(-sest*T);
    end
    end
    
    Ctc = zeros(M,N); %  covariance matrix, C(target,control)
    for i=[1:M]
    for j=[1:N]
        T = abs(tm(i)-td(j));
        Ctc(i,j) = gamma2*cos(w0est*T) * exp(-sest*T);
    end
    end
    
else % use outer product
    % build covariances via outer product
    Tdd = abs( td*ones(1,N) - ones(N,1)*(td') );
    Ccc = gamma2*cos(w0est*Tdd) .* exp(-sest*Tdd);
    Tmd = abs( tm*ones(1,N) - ones(M,1)*(td') );
    Ctc = gamma2*cos(w0est*Tmd) .* exp(-sest*Tmd);
end

% solve GPE problem
A = (Ccc + sigmad2*eye(N,N) );
mest = Ctc * (A\dobs);

% plot mest
plot(tm,mest,'-','LineWidth',2,'Color',[0.8,0.8,0.8]);

% sqrt(variance) of mest

% plot 95% confidence intervals
if( PLOTCONF )
    B = Ctc/A;
    Cttest = sigmad2*B*B';
    sigmamest = sqrt(diag(Cttest));
    plot(tm,mest-2*sigmamest,'k:','LineWidth',2);
    plot(tm,mest+2*sigmamest,'k:','LineWidth',2);
end

% plot estimated covariance
figure(1);
cest = xcorr(mest,mest)/M;
figure(1);
tc = Dtm*[(-M+1):(M-1)]';
plot(tc,cest,'k:','LineWidth',3);


