% edama_09_??, GPR Tuning Example

% this script tunes angular frequency, q, in the
% covariance c(t) = gamma^2 cos(qt) exp(-st)

clear all;

% tunable parameters
MPLOT = 5;    % scale of model parameter plot
CPLOT = 1;    % scale of covariance plot
gamma = 1;    % sqrt(variance) of covariance
sigmad = 0.2; % noise level of observed data
wbias=3.0;    % compute GPR with wrong w0
sbias=1.0;    % compute GPR with wrong s

% axes definitons

% model parameter axis, tm, with sampling Dt
M=256;  % must be even, best if a power of 2
Dtm=1.0;
tmax=Dtm*(M-1); 
tm=Dtm*[0:M-1]'; 

% finely sampled model parameter vector, with sampling
% Dt/K, where K is a smallish power of 2, is used in
% the initialization process, as described below
K=8; % must be even, best if a power of 2
Mf=M*K;  
Dtf=Dtm/K;
tf=Dtf*[0:Mf-1]';

% Part 1: realization of a finely sampled time
% series, mf, with the desired covariance

% PART 1: create true "fine" covariance function

n=6;  % number of oscillations on (0,tmax) interal
w0 = n*2*pi/tmax;  % angular frequency of oscilation
tmid=tmax/2.0;
s = 20.0/tmax;  % decay rate
gamma2 = gamma^2;
cf = gamma2*cos(w0*(tf-tmid)).*exp(-s*abs(tf-tmid));

% plot this covariance
figure(1);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ -tmid, tmid, -CPLOT, CPLOT] );
xlabel('time t (s)');
ylabel('c(t)');
plot(tf-tmid,cf,'k-','LineWidth',3);

% PART 2: create "fine" time series with thus
% covariance using Fourier techniques

% amplitude spectral density of true covariance
cftilde = fft(cf);
cfasd = abs(cftilde);

% To create a time series with a specified covariannce,
% start with a times series of uniformly-distrbuted
% random noise, which has a uniform a.s.d., Fourier
% transorm it, multiply teh transform by the a.s.d
% of the specified covariance, trasnform back, and adjust
% the amplitude to the desired variance.
mf = random('Normal',0.0,1.0,Mf,1); % random numbers
mf = mf - mean(mf); % remove mean
mftilde = fft(mf);      % Fourier transform of mf
mfasd = abs(mftilde);   % a.s.d. of mf
mftilde = (cfasd./mfasd).*mftilde;  % adjust so has a.s.d. of c
mf = real(ifft(mftilde)); % inverse Fourier transform
varmf = var(mf);      % variance of mf
mf = (gamma/sqrt(varmf))*mf; % normalize mf to variance gamma^2

% PART 3: Regularly sub-sample mf to get mtrue, irregularly
% sub-sample mf to get true data dtrue, and add uncorrelated
% noise to dtrue to get observed data dobs

mtrue = mf(1:K:Mf);

N = 50; % fewer observed data than true data
j=zeros(N,1);
Nj=0;
% create list, j, of N unique, randomly chosen indices
% that are not in the sequence 1, K+1, 2K+1, ...
while (Nj<N)
    i=unidrnd(Mf);
    if( (mod(i-1,K)~=0) && (~ismember(i,j)) )
        Nj=Nj+1;
        j(Nj)=i;
    end
end
j = sort(j,'ascend'); % sort for plotting purposes
td = tf(j); % times of observations
dtrue = mf(j);  % irregularly sub-sample mf to get true data
% observed data is true data + uncorrelated random noise
sigmad2 = sigmad^2;
sigmadr = 1.0/sigmad;
dobs = dtrue + random('Normal', 0, sigmad, N, 1 );

% the data kernel, G, for the equation d=Gm
G = [ zeros(N,M), zeros(N,N) ];

% time series figure
figure(2);
clf;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, tmax, -MPLOT, MPLOT] );
xlabel('time t (s)');
ylabel('d(t) and  m(t)');
hold on;
plot(tm,mtrue,'k-','LineWidth',3);
plot(td,dtrue,'ko','LineWidth',2);

% Part 4: GPR solution for inital parameters
% (for comparison purposes)

% Since the true (w0, s) are in general unknown
% this code allows them to be assigned differently
% from the true values
w0est = wbias*w0;  % w0est = 1.0*w0 is the true value
sest = sbias*s;

% build covariances via outer product
Tdd = abs( td*ones(1,N) - ones(N,1)*(td') );
Ccc = gamma2*cos(w0est*Tdd) .* exp(-sest*Tdd);
Tmd = abs( tm*ones(1,N) - ones(M,1)*(td') );
Ctc = gamma2*cos(w0est*Tmd) .* exp(-sest*Tmd);

% solve GPE problem
A = (Ccc + sigmad2*eye(N,N) );
mest = Ctc * (A\dobs);

% plot mest
plot(tm,mest,'-','LineWidth',2,'Color',[0.8,0.8,0.8]);

% plot estimated covariance
figure(1);
cest = xcorr(mest,mest)/M;
figure(1);
tc = Dtm*[(-M+1):(M-1)]';
plot(tc,cest,'k:','LineWidth',3);

% PART 5: Set up for tuning

% covariance plot
figure(3);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ -tmid, tmid, -CPLOT, CPLOT] );
xlabel('time t (s)');
ylabel('c(t)');
plot(tf-tmid,cf,'k-','LineWidth',3);

% time series plot
figure(4);
clf;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, tmax, -MPLOT, MPLOT] );
xlabel('time t (s)');
ylabel('d(t) and  m(t)');
hold on;
plot(tm,mtrue,'k-','LineWidth',3);
plot(td,dtrue,'ko','LineWidth',2);

% Part 6: Tuning

% Newton's Method adjustment of w0, now called q
qtrue=w0;     % true parameter, q
qg = w0est;   % inital guess of parameter, q
Dq = 0.025;   % initial step size for gradiet descent
NITER = 30;   % maximumnumber of iterations
qg = qg-Dq;   % so first iteration starts qg
alpha = 1.0;  % step size reducer
nu = 1.0;     % sign of gradient
mtpri = zeros(M,1); % prior model parameters are zero
mcpri = zeros(N,1); % prior model parameters are zero
mpri = [mtpri;mcpri]; % prior model parameters are zero
tall = [tm; td]; % concatenated time vectors
qsave = zeros(NITER,1); % save estimates, for plot

for iter = [1:NITER]  % loop over descent increments
    
    qg = qg+nu*alpha*Dq; % new guess of parameter, q

    % full covariance matrix and its derivative
    Tall = tall*ones(1,M+N) - ones(M+N,1)*(tall');
    Call = gamma2 * cos(qg*Tall) .* exp(-sest*abs(Tall));
    dCalldq = -gamma2 * Tall .* sin(qg*Tall) .* exp(-sest*abs(Tall));
    
    % break out four quandrants
    Ctt = Call(1:M,1:M);
    dCttdq = dCalldq(1:M,1:M);
    Ctc = Call(1:M,M+1:M+N);
    dCtcdq = dCalldq(1:M,M+1:M+N);
    Cct = Ctc';
    Ccc = Call(M+1:M+N,M+1:M+N);
    dCccdq = dCalldq(M+1:M+N,M+1:M+N);
    
    % A = Ccc + sigmad^2 I
    A = Ccc + sigmad2*eye(N);
    dAdq = dCccdq;
    
    % various functions of the covariance matrix
    CallI = inv(Call);
    CallIsr = sqrtm(CallI);
    dCallIdq = - CallI * dCalldq * CallI;
    dCallIsrdq = sylvester( CallIsr, CallIsr, dCallIdq );
    R = chol(Call);
    logdetCall = sum(log(diag(R)));
    dlogdetCalldq = trace( CallI*dCalldq );

    % GPR solution
    u = A\dobs;
    mtg = Ctc*u;
    mcg = Ccc*u;
    mg = [mtg;mcg];
    dg = mg(M+1:M+N);

    % derivative of predicted data
    v = A\(dAdq*u);
    dmestdq = [dCtcdq*u - Ctc*v; dCccdq*u - Ccc*v];
    ddpredq = dmestdq(M+1:N+M);
    detildedq = -sigmadr*ddpredq;
    e = dobs-dg;
    etilde = sigmadr*e;
    
    % derivative of total data error
    E = etilde'*etilde;
    dEdp = 2*etilde'*detildedq;
    
    % prior error and its derivative
    el=(mpri-mg);
    eltilde = CallIsr*el;
    deltildedq = dCallIsrdq*el - CallIsr*dmestdq;
    
    % total prior error and its derivative
    L = eltilde'*eltilde;
    dLdp = 2*eltilde'*deltildedq;
    
    % function, PSI being minimized
    PSI = logdetCall + E + L;
    % and its deriative
    dPSIdp = dlogdetCalldq + dEdp + dLdp;
    nu = -sign(dPSIdp);  % downhill vector
   
    % gradient method method
    if( iter==1 )  % on the first iteration do nothing but save
        qglast = qg;
        PSIlast = PSI;
        fprintf('1: qg %e nu %e alpha %e PSI %e PSIlast %e\n', qg, nu, alpha, PSI, PSIlast);
    elseif (PSI < PSIlast )
        qglast = qg;
        PSIlast = PSI;
        fprintf('2: qg %e nu %e alpha %e PSI %e PSilast %e\n', qg, nu, alpha, PSI, PSIlast);
    else % psi increased, so decrease alpha
        qg = qglast;
        PSI = PSIlast;
        alpha = alpha/2;
        fprintf('2: qg %e nu %e alpha %e PSI %e PSIlast %e\n', qg, nu, alpha, PSI, PSIlast);
    end
    qsave(iter) = qg;
end
pest = qg;
mest = mg;
mtest = mest(1:M);
mcest = mest(M+1:M+N);
dpre = mcest;

% plot estimated covariance
figure(3);
cest = xcorr(mtest,mtest)/M;
tc = Dtm*[(-M+1):(M-1)]';
plot(tc,cest,'k:','LineWidth',3);

% plot estimated time series
figure(4);
plot( tm, mtest, '-', 'LineWidth', 2, 'Color', [0.8,0.8,0.8]);

fprintf('qtrue %e pest %e relating error %e\n', qtrue, pest, (qtrue-pest)/qtrue );

figure(5);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, NITER, 0, 0.5] );
xlabel('iteration');
ylabel('w0');
plot([1:NITER]',qsave,'k-','LineWidth',2);
plot([1:NITER]',qsave,'ko','LineWidth',2);
plot([0:NITER]',w0*ones(NITER+1,1),'k:','LineWidth',2);



