% Gaussian Process Estimation example
% this example uses synthetic data

clear all;

% standard time set up
N=256;
Dt=1.0;
tmax=Dt*(N-1); 
t=Dt*[0:N-1]'; 

% standard frequency set up
fmax=1/(2.0*Dt);
df=fmax/(N/2);
f=df*[0:N/2,-N/2+1:-1]'; 
Nf=N/2+1;
dw=2*pi*df;
w=dw*[0:N/2,-N/2+1:-1]';
Nw=Nf;

% PART 1: create true covariance function
n=6;
w0 = n*2*pi/tmax;
tmid=tmax/2;
s = 20/tmax;
gamma = 1.0;
gamma2 = gamma^2;
c = gamma2*cos(w0*(t-tmid)).*exp(-s*abs(t-tmid));

% plot true covariance
figure(1);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ -tmid, tmid, -0.5, 1.5] );
xlabel('time t (s)');
ylabel('c(t)');
plot(t-tmid,c,'k-','LineWidth',3);

% amplitude spectral density of true covariance
ct = fft(c);
asd = abs(ct);

% plot amplitude spectral density
figure(2);
clf;
hold on;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, fmax, 0, 15] );
xlabel('frequency f (Hz)');
ylabel('c(f)');
plot(f,asd,'k-','LineWidth',3);

% PART 2: create synthetic data with this covariance

% realize a time series d with this covariance 
d = random('normal',0,1,N,1); % random numbers
d = d - mean(d); % remove mean
dt = fft(d);     % Fourier transform of d
dasd = abs(dt);  % a.s.d. of d
dt = asd.*dt./dasd; % adjust so has a.s.d. of c
dtrue = real(ifft(dt)); % inverse Fourier transform
vard = var(dtrue);      % variance of dtrue
dtrue = gamma*dtrue/sqrt(vard); % normalize d to variance gamma^2
dasd = abs(fft(dtrue)); % corrected a.s.d of dtrue

% time series figure
figure(3);
clf;
set(gca,'LineWidth',2);
set(gca,'FontSize', 14);
axis( [ 0, N, -5, 5] );
xlabel('time t (s)');
ylabel('d(t) and  m(t)');
hold on;
plot(t,dtrue,'k-','LineWidth',3);

% PART 3: Create observed data by sub-sampling true
% data and adding random noise

% noise level
sigma = 0.5;
sigma2 = sigma^2;

Nd = 50; % fewer observed data than true dat
j = unidrnd( N, Nd, 1 ); % randomly select times
j = sort(j,'ascend'); % sort for plotting purposes
tobs = t(j); % observed times
% observed data is true data + uncorrelated random noise
dobs = dtrue(j) + random('Normal', 0, sigma, Nd, 1 );

% plot observed data
figure(3);
plot(tobs,dobs,'ko','LineWidth',2);

% PART 4: Gaussian Process Estimation

% assign w0 and s used in GPR
w0est = 2.0*w0; % w0est = 1.0*w0 is the true value
sest = 1.0*s;

% Build covariance C(control,control)
C1 = zeros(Nd,Nd);
for i=[1:Nd]
for j=[1:Nd]
    C1(i,j) = gamma2*cos(w0est*(tobs(i)-tobs(j))) .* exp( -sest*abs(tobs(i)-tobs(j)) );
end
end

% Build covariance C(target,control)
C2 = zeros(N,Nd);
for i=[1:N]
for j=[1:Nd]
    C2(i,j) = gamma2*cos(w0est*(t(i)-tobs(j))) .* exp( -sest*abs(t(i)-tobs(j)) );
end
end

% solve GPE problem
M = (C1 + sigma2*eye(Nd,Nd) );
mest = C2 * (M\dobs);

% plot m = estimated target data
figure(3);
plot(t,mest,'k:','LineWidth',3);

% a.s.d. of estimated target data
mt = fft(mest);
masd = abs(mt);

% plot a.s.d. of estimated target data
figure(2);
norm=2/sqrt(Nf);
plot(f(1:Nf),masd(1:Nf)*norm,'k-','LineWidth',2);

% compute covariance of estimated target data
cm = real(ifft((masd.*2).*exp(complex(0,1)*w*tmid-Dt)));
figure(1);
norm=1/4;
plot(t-tmid,cm*norm,'k:','LineWidth',3);

